/*
 *  Bluelog - Fast Bluetooth scanner with optional Web frontend
 * 
 *  Bluelog is a Bluetooth site survey tool, designed to tell you how
 *  many discoverable devices there are in an area as quickly as possible.
 *  As the name implies, its primary function is to log discovered devices
 *  to file rather than to be used interactively. Bluelog could run on a
 *  system unattended for long periods of time to collect data. In addition
 *  to basic scanning, Bluelog also has a unique feature called "Bluelog Live",
 *  which puts results in a constantly updating Web page which you can serve
 *  with your HTTP daemon of choice.
 * 
 *  Bluelog uses code from a number of GPL projects. See README for more info.
 *
 *  Written by Tom Nardi (MS3FGX@gmail.com), released under the GPLv2.
 *  For more information, see: www.digifail.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <syslog.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>

#include "classes.c"

#define VERSION	"0.9.8"
#define APPNAME "Bluelog"
#define PID_FILE "/tmp/bluelog.pid"

// Global variables
FILE *outfile; // Output file
FILE *infofile; // Status file
inquiry_info *results; // BlueZ scan results struct
char **dev_cache; // Device cache
int bt_socket; // HCI device
int showtime = 0; // Show timestamps in log
						
char* get_time()
{
	// Time variables
	time_t rawtime;
	struct tm * timeinfo;
	static char time_string[20];
	
	// Find time and put it into time_string
	time (&rawtime);
	timeinfo = localtime(&rawtime);
	strftime(time_string,20,"%D %T",timeinfo);
	
	// Send it back
	return(time_string);
}

void shut_down(int sig)
{
	// Close up shop
	printf("\n");
	printf("Closing files and freeing memory...");
	// Only show this if timestamps are enabled
	if (showtime)
		fprintf(outfile,"[%s] Scan ended.\n", get_time());
	
	fclose(outfile);
	free(results);
	free(dev_cache);
	close(bt_socket);
	printf("OK\n");
	printf("Done!\n");
	// Log shutdown to syslog
	syslog(LOG_INFO, "Shutdown OK");
	// Delete PID file
	unlink(PID_FILE);
	exit(sig);
}

void live_entry (char *timefound, char *addr, char *name, char *class, char *capabilities)
{
	// Let's format these a little nicer
	if (strcmp(name, "VOID") == 0)
		name = "No Response";
	if (strcmp(class, "VOID") == 0)
		class = "Unclassified";
	if (strcmp(capabilities, "VOID") == 0)
		capabilities = "Not Reported";
		
	// Write out HTML code
	fprintf(outfile,"<tr>");
	fprintf(outfile,"<td>%s</td>", timefound);
	fprintf(outfile,"<td>%s</td>", addr);
	fprintf(outfile,"<td>%s</td>", name);
	fprintf(outfile,"<td>%s</td>", class);
	fprintf(outfile,"<td>%s</td>", capabilities);
	fprintf(outfile,"</tr>\n");
}

int read_pid (void)
{
	// Any error will return 0
	FILE *pid_file;
	int pid;

	if (!(pid_file=fopen(PID_FILE,"r")))
		return 0;
		
	if (fscanf(pid_file,"%d", &pid) < 0)
		pid = 0;

	fclose(pid_file);	
	return pid;
}

static void write_pid (pid_t pid)
{
	FILE *pid_file;
	
	// Open PID file
	printf("Writing PID file: %s...", PID_FILE);
	if ((pid_file = fopen(PID_FILE,"w")) == NULL)
	{
		printf("\n");
		printf("Error opening PID file!\n");
		exit(1);
	}
	printf("OK\n");
	
	// If open, write PID and close	
	fprintf(pid_file,"%d\n", pid);
	fclose(pid_file);
}

static void daemonize (void)
{
	// Process and Session ID
	pid_t pid, sid;
	
	syslog(LOG_INFO,"Starting as daemon...");
 
	// Fork off process
	pid = fork();
	if (pid < 0)
		exit(EXIT_FAILURE);
	else if (pid > 0)
		exit(EXIT_SUCCESS);
			
	// Change umask
	umask(0);
 
	// Create a new SID for the child process
	sid = setsid();
	if (sid < 0)
		exit(EXIT_FAILURE);

	// Change current working directory
	if ((chdir("/")) < 0)
		exit(EXIT_FAILURE);
		
	// Write PID file
	write_pid(sid);
	
	printf("Going into background...\n");
		
	// Close file descriptors
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);
}

static void help(void)
{
	printf("%s (v%s) by MS3FGX\n", APPNAME, VERSION);
	printf("----------------------------------------------------------------\n");
	printf("Bluelog is a Bluetooth site survey tool, designed to tell you how\n"
		"many discoverable devices there are in an area as quickly as possible.\n"
		"As the name implies, its primary function is to log discovered devices\n"
		"to file rather than to be used interactively. Bluelog could run on a\n"
		"system unattended for long periods of time to collect data.\n");
	printf("\n");
	printf("Bluelog also includes a mode called \"Bluelog Live\" which creates a\n"
		"webpage of the results which you can serve up with your HTTP daemon of\n"
		"choice.\n");
	printf("\n");
	printf("For more information, see www.digifail.com\n");
	printf("\n");
	printf("Options:\n"
		"\t-i <interface>     Sets scanning device, default is hci0\n"
		"\t-o <filename>      Sets output filename, default is devices.log\n"
		"\t-r <attempts>      Name resolution retries, default is 3\n"
		"\t-w <milliseconds>  Name resolution retry delay, default is 500 ms\n"
		"\t-b                 Enable BlueProPro log format, see README\n"
		"\t-c                 Write device class to log, default is disabled\n"
		"\t-d                 Daemonize\n"
		"\t-f                 Use \"friendly\" device class, default is disabled\n"
		"\t-k                 Kill an already running Bluelog process\n"
		"\t-l                 Start \"Bluelog Live\", default is disabled\n"
		"\t-t                 Write timestamps to log, default is disabled\n"
		"\t-v                 Enables verbose output, default is disabled\n"
		"\t-x                 Obfuscate discovered MACs, default is disabled\n"
		"\n");
}

static struct option main_options[] = {
	{ "interface", 1, 0, 'i' },
	{ "output",	1, 0, 'o' },
	{ "verbose", 0, 0, 'v' },
	{ "retry", 1, 0, 'r' },
	{ "wait", 1,0, 'w' },
	{ "time", 0, 0, 't' },
	{ "obfuscate", 0, 0, 'x' },
	{ "class", 0, 0, 'c' },
	{ "live", 0, 0, 'l' },
	{ "kill", 0, 0, 'k' },
	{ "friendly", 0, 0, 'f' },
	{ "bluepropro", 0, 0, 'b' },
	{ "help", 0, 0, 'h' },
	{ "daemonize", 0, 0, 'd' },
	{ 0, 0, 0, 0 }
};

int main(int argc, char *argv[])
{	
	// Handle signals
	signal(SIGINT,shut_down);
	signal(SIGHUP,shut_down);
	signal(SIGTERM,shut_down);
	signal(SIGINT,shut_down);
	signal(SIGQUIT,shut_down);

	// HCI device number, MAC struct
	int device = 0;
	bdaddr_t bdaddr;
	bacpy(&bdaddr, BDADDR_ANY);
	
	// Time to scan. Scan time is roughly 1.28 seconds * scan_time
	int scan_time = 8;
	
	// Maximum number of devices per scan
	int max_results = 255;
	int num_results;
	
	// Device cache maximum and index
	int max_cache = 512;
	int cache_index = 0;

	// HCI cache setting
	int flags = IREQ_CACHE_FLUSH;
	
	// Strings to hold MAC and name
	char addr[19] = {0};
	char name[248] = {0};
	
	// Variables for name routine
	int nametry;
	int retry_count = 3;
	int timeout = 0;
	int delay = 500;
	
	// String for time
	char cur_time[20];
	
	// Process ID read from PID file
	int ext_pid;
	
	// Settings
	int verbose = 0;
	int obfuscate = 0;
	int showclass = 0;
	int friendlyclass =0;
	int bluelive = 0;
	int daemon = 0;
	int bluepropro = 0;
	
	// Pointers to filenames, with default output filename.
	char *outfilename = "devices.log";
	char *infofilename = "/tmp/info.txt";
	
	// Mode to open output file in
	char *filemode = "a+";
	
	// Misc Variables
	int i, ri, oi, opt;
	
	while ((opt=getopt_long(argc,argv,"+o:i:r:w:vxcthldbfk", main_options, NULL)) != EOF)
	{
		switch (opt)
		{
		case 'i':
			if (!strncasecmp(optarg, "hci", 3))
				hci_devba(atoi(optarg + 3), &bdaddr);
			else
				str2ba(optarg, &bdaddr);
			break;
		case 'o':
			outfilename = strdup(optarg);
			break;
		case 'r':
			retry_count = atoi(optarg);
			break;
		case 'w':
			delay = atoi(optarg);
			break;
		case 'c':
			showclass = 1;
			break;
		case 'f':
			friendlyclass = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		case 't':
			showtime = 1;
			break;
		case 'x':
			obfuscate = 1;
			break;
		case 'l':
			bluelive = 1;
			break;
		case 'b':
			bluepropro = 1;
			break;
		case 'd':
			daemon = 1;
			break;
		case 'h':
			help();
			exit(0);
		case 'k':
			// Read PID from file into variable
			ext_pid = read_pid();
			if (ext_pid != 0)
			{
				printf("Killing Bluelog process with PID %i...",ext_pid);
				kill(ext_pid,15);
				printf("OK.\n");
				
				// Delete PID file
				unlink(PID_FILE);
			}
			else
				printf("No running Bluelog process found.\n");
				
			exit(0);
		default:
			printf("Unknown option. Use -h for help, or see README.\n");
			exit(1);
		}
	}
	
	// See if there is already a process running
	if (read_pid() != 0)
	{
		printf("Another instance of Bluelog is already running!\n");
		printf("Use the -k option to kill a running Bluelog process.\n");
		exit(1);
	}
	
	// Sanity checks
	if (retry_count < 0 || delay < 0)
	{	
		printf("Error, arguments must be positive numbers!\n");
		exit(1);
	}
	
	// Override some options that don't play nice with others
	// No verbose for daemon
	if (daemon)
		verbose = 0;
		
	// No Bluelog Live when running BPP
	if (bluepropro)
		bluelive = 0;	

	// Showing raw class ID turns off friendly names
	if (showclass)
		friendlyclass = 0;
			
	// No timestamps for Bluelog Live
	if (bluelive)
		showtime = 0;	
		
	// Boilerplate
	printf("%s (v%s) by MS3FGX\n", APPNAME, VERSION);
	printf("---------------------------\n");

	// Init Hardware
	ba2str(&bdaddr, addr);
	if (!strcmp(addr, "00:00:00:00:00:00"))
	{
		printf("Autodetecting device...");
		device = hci_get_route(NULL);
		// Put autodetected device MAC into addr
		hci_devba(device, &bdaddr);
		ba2str(&bdaddr, addr);
	}
	else
	{
		printf("Initializing device...");
		device = hci_devid(addr);
	}
	
	// Open device and catch errors
	bt_socket = hci_open_dev(device); 
	if (device < 0 || bt_socket < 0)
	{
		// Failed to open device, that can't be good
		printf("ERROR!\n");
		exit(1);
	}
	// If we get here the device should be online.
	printf("OK\n");
	
	// Allocate result cache
	printf("Allocating result cache...");
	dev_cache = (char**)calloc(max_cache,sizeof(char**));
 	if (dev_cache == NULL)
 	{
		printf("Error allocating device cache!\n");
 		exit(1);
 	}
 	printf("OK\n");
	
	// Open output file
	if (bluelive)
	{
		// Change location of output file
		outfilename = "/tmp/devices.log";
		filemode = "w";
		printf("Starting Bluelog Live...\n");
	}
	printf("Opening output file: %s...", outfilename);
	if ((outfile = fopen(outfilename, filemode)) == NULL)
	{
		printf("\n");
		printf("Error opening output file!\n");
		exit(1);
	}
	printf("OK\n");
	
	// Open status file
	if (bluelive)
	{
		printf("Opening info file: %s...", infofilename);
		if ((infofile = fopen(infofilename,"w")) == NULL)
		{
			printf("\n");
			printf("Error opening info file!\n");
			exit(1);
		}
		printf("OK\n");
	}

	// Write PID file
	if (!daemon)
		write_pid(getpid());
	
	// Get and print time to console and file
	strcpy(cur_time, get_time());
	if (!daemon)
		printf("Scan started at [%s] on %s.\n", cur_time, addr);
	if (showtime)
	{
		fprintf(outfile,"[%s] Scan started on %s\n", cur_time, addr);
		// Make sure this gets written out
		fflush(outfile);
	}
		
	// Write info file for Bluelog Live
	if (bluelive)
	{
		fprintf(infofile,"<div class=\"sideitem\">%s Version: %s</div>\n", APPNAME, VERSION);
		fprintf(infofile,"<div class=\"sideitem\">Device: %s</div>\n", addr);
		fprintf(infofile,"<div class=\"sideitem\">Started: %s</div>\n", cur_time);
		
		// Think we are done with you now
		fclose(infofile);
	}
		
	// Log success to this point
	syslog(LOG_INFO,"Init OK!");
	
	// Daemon switch
	if (daemon)
		daemonize();
	else
		printf("Hit Ctrl+C to end scan.\n");
	
	// Start scan, be careful with this infinite loop...
	for(;;)
	{	
		// Init result struct
		results = (inquiry_info*)malloc(max_results * sizeof(inquiry_info));
		// Scan and return number of results
		num_results = hci_inquiry(device, scan_time, max_results,
		NULL, &results, flags);
		if(num_results < 0)
		{
			// If returned results is less than 0, exit with failure
			printf("Scan failed!\n");
			shut_down(1);
		}

		// Check if we need to reset device cache
		if (cache_index == max_cache)
		{
			syslog(LOG_INFO,"Resetting device cache...");
			for (ri = 0; ri <= max_cache; ri++)
				dev_cache[ri] = NULL;
		}
		
		// Loop through results
		for (i = 0; i < num_results; i++)
		{
			// Return MAC from struct
			ba2str(&(results+i)->bdaddr, addr);

			// Compare to device cache
			for (ri = 0; ri <= cache_index; ri++)
			{
						
				if (dev_cache[ri] != NULL) // Previous found device
				{
					if (strcmp(addr, dev_cache[ri]) == 0) // Match found, exit
					{
						break;
					}
				}
				else if (dev_cache[ri] == NULL) // New device found
				{			
					// Write new device to cache
					// Allocate size of addr, copy string (NOT pointer)
					dev_cache[cache_index] = (char *) calloc (19, sizeof (char));
					strcpy(dev_cache[cache_index], addr);
					cache_index++;
					
					// Find device name
					// We will try three times to get a response
					memset(name,'\0', sizeof(name));
					for (nametry = 1; nametry <= retry_count; nametry++)
					{		
						// If we get a response, exit out of loop
						if (hci_read_remote_name(bt_socket, &(results+i)->bdaddr,\
							sizeof(name), name, timeout) == 0)
							break;
						
						// Don't really want to do this...
						usleep(delay);
					}
					// If name is still blank at this point, make default
					if (strcmp(name,"") == 0)
						strcpy(name,"VOID");
					
					// Get time again
					strcpy(cur_time, get_time());
					
					// Obfuscate last octet of MAC
					if (obfuscate)
						for (oi = 15; oi <= 16; oi++)
							addr[oi] = 'X';
					
					// Write to syslog if we are daemon
					if (daemon)
						syslog(LOG_INFO,"Found new device: %s",addr);
					
					// Print everything to console if verbose is on
					if (verbose)
						printf("[%s] %s,%s,0x%02x%02x%02x\n", cur_time,\
						addr, name, (results+i)->dev_class[2],\
						(results+i)->dev_class[1], (results+i)->dev_class[0]);
											
					if (bluelive)
					{
						// Write result with live function
						live_entry(cur_time, addr, name,\
						device_class((results+i)->dev_class),\
						device_capability((results+i)->dev_class));
					}
					else if (bluepropro)
					{
						// Set output format for BlueProPro
						fprintf(outfile,"%s,%s", name, addr);
						fprintf(outfile,",0x%02x%02x%02x\n",\
						(results+i)->dev_class[2],\
						(results+i)->dev_class[1],\
						(results+i)->dev_class[0]);
					}
					else 
					{
						// Print time first if enabled
						if (showtime)
							fprintf(outfile,"[%s] ", cur_time);
							
						// Always output MAC
						fprintf(outfile,"%s", addr);
						
						// Optionally output class
						if (showclass)					
							fprintf(outfile,",0x%02x%02x%02x",\
							(results+i)->dev_class[2], (results+i)->dev_class[1],\
							(results+i)->dev_class[0]);
							
						// "Friendly" version of class info
						if (friendlyclass)					
							fprintf(outfile,",%s,(%s)",\
							device_class((results+i)->dev_class),\
							device_capability((results+i)->dev_class));
							
						// Append the name
						fprintf(outfile,",%s", name);
						
						// Close with newline
						fprintf(outfile,"\n");
					}
					break;
				}
			}
			// Write any new changes to output file
			fflush(outfile);
		}
	}
	// If we get here, shut down
	shut_down(0);
	// STFU
	return (1);
}

